@@ -72,31 +72,86 @@ def merge_dict(d1, d2):
     return d1
 
 
+class Node(object):
+    def __init__(self, name, deps=[], fdto={}, tool=None):
+        self.deps = deps
+        self.fdto = fdto
+        self.tool = tool
+
+        # Import and instantiate the tool class requested by "tool"
+        self.inst = getattr(import_module(f"edalize.tools.{tool}"), tool.capitalize())()
+
+
+class FlowGraph(object):
+    def __init__(self):
+        self._graph = {}
+
+    @classmethod
+    def fromdict(cls, d):
+        c = FlowGraph()
+        _d = d.copy()
+        while _d:
+            # Make a copy to avoid popping from the dict that is being iterated over
+            # and to compare length afterwards. If length is same, it means no item
+            # was popped and graph is not satisifiable
+            _d2 = _d.copy()
+            for k, v in _d2.items():
+
+                # It is safe to pop the element if all dependencies of the node
+                # exist in the graph already
+                if set(v.get("deps", [])) <= set(c.get_nodes()):
+                    node = _d.pop(k)
+
+                    # node["deps"] is a list of strings.
+                    # Get the corresponding node objects for each entry
+                    deps = [c.get_node(dep) for dep in node.get("deps", [])]
+
+                    # tool name is the key by default
+                    tool = v.get("tool", k)
+
+                    c._graph[k] = Node(
+                        k, deps=deps, fdto=node.get("fdto", {}), tool=tool
+                    )
+            if len(_d) == len(_d2):
+                raise RuntimeError("Unsatisfiable graph")
+        return c
+
+    def add_node(self, name, node):
+        self._graph[name] = node
+
+    def get_node(self, name):
+        return self._graph[name]
+
+    def get_nodes(self):
+        return self._graph
+
+
 class Edaflow(object):
 
-    FLOW = []
+    FLOW_OPTIONS = {}
 
     @classmethod
     def get_flow_options(cls):
-        flow_opts = cls.FLOW_OPTIONS.copy()
-        for tool in cls.FLOW:
-            (tool_name, next_nodes, flow_defined_tool_options) = tool
-
-            # Get available tool options from each tool in the flow
-            class_tool_options = getattr(
-                import_module(f"edalize.tools.{tool_name}"), tool_name.capitalize()
-            ).get_tool_options()
-            for opt_name in class_tool_options:
-                # Filter out tool options that are already set by the flow
-                if not opt_name in flow_defined_tool_options:
-                    flow_opts[opt_name] = class_tool_options[opt_name]
-                    flow_opts[opt_name]["tool"] = tool_name
-        return flow_opts
+        return cls.FLOW_OPTIONS.copy()
 
     @classmethod
     def get_tool_options(cls, flow_options):
         return {}
 
+    @classmethod
+    def _require_flow_option(cls, flow_options, option_name):
+        """Check for mandatory flow option.
+
+        Returns the value if it exists or otherwise throws a RuntimeError
+        """
+
+        option = flow_options.get(option_name, "")
+        if not option:
+            raise RuntimeError(
+                f"Flow '{cls.__name__.lower()}' requires flow option '{option_name}' to be set"
+            )
+        return option
+
     # Takes a list of tool names and a dict of pre-defined tool options
     # Imports the tool class for every tool in the list, extracts their
     # tool options and return then all, except for the ones listed in
@@ -124,79 +179,54 @@ class Edaflow(object):
         return tool_opts
 
     def extract_flow_options(self):
-        # Extract flow options from the EDAM
-        flow_options = {}
-        available_flow_options = [
-            k for k, v in self.get_flow_options().items() if not v.get("tool")
-        ]
-        edam_flow_opts = self.edam.get("flow_options", {})
-        for opt_name in list(edam_flow_opts.keys()):
-            if opt_name in available_flow_options:  # self.get_flow_options():
-                flow_options[opt_name] = edam_flow_opts.pop(opt_name)
-        return flow_options
+        return {
+            k: v
+            for (k, v) in self.edam.get("flow_options", {}).items()
+            if k in self.get_flow_options()
+        }
 
     # Filter out tool options for each tool from self.flow_options
     def extract_tool_options(self):
         tool_options = {}
         edam_flow_opts = self.edam.get("flow_options", {})
-        for (tool_name, next_nodes, flow_defined_tool_options) in self.flow:
-
+        for name, node in self.flow.get_nodes().items():
             # Get the tool class
             ToolClass = getattr(
-                import_module(f"edalize.tools.{tool_name}"), tool_name.capitalize()
+                import_module(f"edalize.tools.{node.tool}"), node.tool.capitalize()
             )
             # Inject the flow-defined tool options to the EDAM
-            tool_options[tool_name] = merge_dict(
-                flow_defined_tool_options, tool_options.get(tool_name, {})
+            tool_options[node.tool] = merge_dict(
+                node.fdto, tool_options.get(node.tool, {})
             )
 
             # Assign the EDAM-defined tool options to the right tool
             for opt_name in list(edam_flow_opts.keys()):
                 if opt_name in ToolClass.get_tool_options():
-                    tool_options[tool_name] = merge_dict(
-                        tool_options[tool_name],
+                    tool_options[node.tool] = merge_dict(
+                        tool_options[node.tool],
                         {opt_name: edam_flow_opts.get(opt_name)},
                     )
 
             self.edam["tool_options"] = tool_options
 
-    def build_tool_graph(self):
-        # Instantiate the tools
-        nodes = {}
-        for (tool_name, next_nodes, flow_defined_tool_options) in self.flow:
-
-            # Instantiate the tool class
-            tool_inst = getattr(
-                import_module(f"edalize.tools.{tool_name}"), tool_name.capitalize()
-            )()
-
-            # FIXME: Don't like injecting stuff like this
-            tool_inst.next_nodes = next_nodes
-            tool_inst.work_root = self.work_root
-
-            nodes[tool_name] = tool_inst
-
-        for name, node in nodes.items():
-            for next_node in node.next_nodes:
-                # Add backwards references
-                nodes[next_node].prev_nodes.add(node)
-        return nodes
-
-    def configure_tools(self, nodes):
+    def configure_tools(self, graph):
         def merge_edam(a, b):
             # Yeah, I know. It's just a temporary hack
             return b
 
-        unconfigured_nodes = list(nodes.values())
+        # Instantiate each node and add to list of unconfigured nodes
+        unconfigured_nodes = list(graph.get_nodes().values())
+
+        # Configure each node in graph order
         while unconfigured_nodes:
             node = unconfigured_nodes.pop(0)
             input_edam = {}
 
             # Check all dependencies are fulfilled
             all_deps_configured = True
-            for n in node.prev_nodes:
-                if n.edam:
-                    input_edam = merge_edam(input_edam, n.edam)
+            for n in node.deps:
+                if n.inst.edam:
+                    input_edam = merge_edam(input_edam, n.inst.edam)
                 else:
                     all_deps_configured = False
 
@@ -206,16 +236,17 @@ class Edaflow(object):
                 if not input_edam:
                     input_edam = self.edam
 
-                node.configure(input_edam)
+                node.inst.work_root = self.work_root
+                node.inst.setup(input_edam)
 
                 # This is an input node. Inject dependency on pre_build scripts
-                if not node.prev_nodes:
+                if not node.deps:
                     # Inject pre-build scripts before the first command
                     # that the node executes. Note that this isn't
                     # technically correct since the first command in
                     # the list might not be the first command executed
-                    node.commands[0].order_only_deps = ["pre_build"]
-                self.commands.commands += node.commands
+                    node.inst.commands.commands[0].order_only_deps = ["pre_build"]
+                self.commands.commands += node.inst.commands.commands
 
     def add_scripts(self, depends, hook_name):
         last_script = depends
@@ -234,9 +265,6 @@ class Edaflow(object):
             last_script = script["name"]
         self.commands.add([], [hook_name], [last_script])
 
-    def configure_flow(self, flow_options):
-        return self.FLOW
-
     def __init__(self, edam, work_root, verbose=False):
         self.edam = edam
         self.commands = EdaCommands()
@@ -256,19 +284,11 @@ class Edaflow(object):
         self.stdout = None
         self.stderr = None
 
-    def set_run_command(self):
-        self.commands.add([], ["run"], ["pre_run"])
-
-    def configure(self):
-
         # Add pre build hooks
         self.add_scripts("", "pre_build")
 
-        # Instantiate all tools (nodes) and build a DAG of the flow
-        nodes = self.build_tool_graph()
-
         # Configure the individual tools in the graph
-        self.configure_tools(nodes)
+        self.configure_tools(self.flow)
 
         # Add post_build scripts to the end of the build chain
         self.add_scripts(self.commands.default_target, "post_build")
@@ -279,10 +299,19 @@ class Edaflow(object):
         self.set_run_command()
         self.add_scripts("run", "post_run")
 
+    def set_run_command(self):
+        self.commands.add([], ["run"], ["pre_run"])
+
+    def configure(self):
+
+        # Write tool-specific config files
+        for node in self.flow.get_nodes().values():
+            node.inst.configure()
+
         # Write out execution file
         self.commands.write(os.path.join(self.work_root, "Makefile"))
 
-    def _run_tool(self, cmd, args=[], cwd=None, quiet=False):
+    def _run_tool(self, cmd, args=[], cwd=None, quiet=False, env={}):
         logger.debug("Running " + cmd)
         logger.debug("args  : " + " ".join(args))
 
@@ -296,6 +325,7 @@ class Edaflow(object):
                 stderr=self.stderr,
                 capture_output=capture_output,
                 check=True,
+                env={**os.environ, **env},
             )
         except FileNotFoundError:
             _s = "Command '{}' not found. Make sure it is in $PATH".format(cmd)
@@ -319,5 +349,5 @@ class Edaflow(object):
         self._run_tool("make", cwd=self.work_root)
 
     # Most flows won't have a run phase
-    def run(self, args):
+    def run(self, args=None):
         pass
